use std::io::{Read, Write};
use std::{env, thread};
use std::error::Error;
use std::net::{Shutdown, TcpListener, TcpStream};
use std::sync::{Arc, Mutex};

use rcgen::generate_simple_self_signed;
use rustls::crypto::aws_lc_rs::sign::any_supported_type;
use rustls::pki_types::pem::PemObject;
use rustls::pki_types::PrivateKeyDer;
use rustls::server::ResolvesServerCertUsingSni;
use rustls::sign::CertifiedKey;
use rustls::{ServerConfig, ServerConnection, Stream};
use threadpool::ThreadPool;

fn accept_client(stream: &mut TcpStream, config: Arc<ServerConfig>, sni_domain: String, host: String) -> Result<(), Box<dyn Error>> {
    let remote_stream = Arc::new(TcpStream::connect(host)?);

    let connection = ServerConnection::new(config)?;
    let connection = Arc::new(Mutex::new(connection));
    
    {
        let mut conn = connection.lock().unwrap();
        let mut tls_stream = Stream::new(&mut *conn, stream);
        while tls_stream.conn.is_handshaking() {
            tls_stream.conn.complete_io(&mut tls_stream.sock)?;
        }

        if tls_stream.conn.server_name() != Some(&sni_domain) {
            return Err("unexpected sni domain".into());
        }
    }

    thread::spawn({
        let mut stream = stream.try_clone()?;
        let connection = connection.clone();
        let remote_stream = remote_stream.clone();

        move || {
            loop {
                let mut buffer = vec![0; 4096];
                let n = (&*remote_stream).read(&mut buffer).unwrap();
                if n != 0 {
                    buffer.truncate(n);

                    let mut conn = connection.lock().unwrap();
                    let mut tls_stream = Stream::new(&mut *conn, &mut stream);
                    tls_stream.write_all(&buffer[..n]).unwrap();
                    tls_stream.flush().unwrap();
                }
            }
        }
    });

    let mut stream = stream.try_clone()?;
    let connection = connection.clone();
    
    loop {
        let mut conn = connection.lock().unwrap();
        let mut tls_stream = Stream::new(&mut *conn, &mut stream);

        let mut buffer = vec![0; 4096];
        let n = tls_stream.read(&mut buffer)?;
        if n != 0 {
            buffer.truncate(n);
            (&*remote_stream).write(&buffer)?;
        }
    }
}

fn main() -> Result<(), Box<dyn Error>> {
    let mut args = env::args();

    args.next();

    let sni_domain = args
        .next()
        .expect("missing sni domain argument");
    let host = args
        .next()
        .expect("missing host argument");
    let local_host = args
        .next()
        .expect("missing local host argument");

    let certified_key = generate_simple_self_signed(vec![sni_domain.to_string()]).unwrap();
    let certified_key = CertifiedKey::new(
        vec![certified_key.cert.into()], 
        any_supported_type(
            &PrivateKeyDer::from_pem_slice(certified_key.key_pair.serialize_pem().as_bytes())?
        )?
    );

    let mut resolver = ResolvesServerCertUsingSni::new();
    resolver.add(&sni_domain, certified_key)?;

    let config = Arc::new(ServerConfig::builder()
        .with_no_client_auth()
        .with_cert_resolver(Arc::new(resolver)));

    let threadpool = ThreadPool::new(10);

    let listener = TcpListener::bind(local_host).unwrap();

    for stream in listener.incoming() {
        let mut stream = stream.expect("listener got broken");

        threadpool.execute({
            let config = config.clone();
            let sni_domain = sni_domain.clone();
            let host = host.clone();

            move || {
                if let Err(e) = accept_client(&mut stream, config, sni_domain, host) {
                    println!("error connection: {e:?}");
                }

                let _ = stream.shutdown(Shutdown::Both);
            }
        });
    }

    Ok(())
}